// -*-c++-*-
//===========================================================
//=     University of Illinois at Urbana/Champaign          =
//=     Department of Computer Science                      =
//=     Dr. Dan Roth - Cognitive Computation Group          =
//=                                                         =
//=  Project: Snow                                          =
//=                                                         =
//=   Module: Cloud.h                                       =
//=  Version: 2.1                                           =
//=  Authors: Jeff Rosen, Andrew Carlson                    =
//=     Date: xx/xx/99                                      = 
//=                                                         =
//= Comments:                                               =
//===========================================================

#ifndef __CLOUD_H__
#define __CLOUD_H__

#include "Target.h"
#include "TargetIdSet.h"
#include "TargetRank.h"
#include "NaiveBayes.h"
#include "GlobalParams.h"
#include <vector>
#include <numeric>
#include <iomanip>

using namespace std;

typedef vector<LearningAlgorithm*> AlgorithmVector;

class Cloud
{
public:
    Cloud();
    Cloud( FeatureID featid);
    Cloud( FeatureID featid, AlgorithmVector& algorithms );
    
    bool operator==( const Cloud& rhs ) const;
    bool operator!=( const Cloud& rhs ) const;
    bool operator<( const Cloud& rhs ) const;
    bool operator<=( const Cloud& rhs ) const;
    bool operator>( const Cloud& rhs ) const;
    bool operator>=( const Cloud& rhs ) const;
	
    // Functions used during training
    bool PresentExample( Example& ex );
    void TrainingComplete();
    void NormalizeConfidence();
    void Discard();
    void Discard( FeatureID id, double delta );

    // Functions used during evaluation and on-line learning
    double RankTargets( Example& ex);
    void PreparetoRank( Example& ex);
    void FindMinimumWeight( FeatureID id, double& min, int& dups);
    void CollectFeatures( set<FeatureID>& featureSet );
    void ResetCounters();
    void ShowStatistics( Counter total );
    
    // Output functions
    void Show();
    void WriteAlgorithms( ofstream& out );
    FeatureID Id() const;   

    // Manual cloud construction
    void AddTarget( const Target& target );
        
    // Network persistence
    void Write( ofstream& out );

protected:
    typedef vector<Target> TargetVector;
    
    FeatureID       targetID;
    TargetVector    targets;
};
 
inline Cloud::Cloud() 
{
}

inline Cloud::Cloud( FeatureID featid )
{
    targetID = featid;
}

inline Cloud::Cloud( FeatureID featid, AlgorithmVector& algorithms ) 
{
    AlgorithmVector::iterator algIt = algorithms.begin();
    AlgorithmVector::iterator algEnd = algorithms.end();
  
    // Create one new Target for each learner 
    for ( ; algIt != algEnd; algIt++)    
    {
        targets.push_back(Target(featid, *algIt));
    }
  
    targetID = featid;  
}

inline void Cloud::AddTarget( const Target& tar )
{
    targets.push_back(tar);
}

inline bool Cloud::PresentExample( Example& ex )
{
    int mistakes = 0;
    
    // Walk through targets and present the example
    TargetVector::iterator it = targets.begin();
    TargetVector::iterator end = targets.end();
    for ( ; it != end; it++)
    {
        mistakes += it->PresentExample(ex) ? 1 : 0;
    }
    
    return (mistakes > 0);    
	
}

inline void Cloud::TrainingComplete()
{
    TargetVector::iterator it = targets.begin();
    TargetVector::iterator end = targets.end();
    
    for ( ; it != end; it++)
    {   
        it->TrainingComplete();
    }
    NormalizeConfidence();
}

inline void Cloud::NormalizeConfidence()
{
    double totalConfidence = 0;

    TargetVector::iterator it = targets.begin();
    TargetVector::iterator end = targets.end();

    for ( ; it != end; it++)
    {   
        totalConfidence += it->Confidence();
    }   
    for ( it = targets.begin(); it != end; it++)
    {
        it->NormalizeConfidence(totalConfidence);
    }
}

inline void Cloud::Discard()
{
	TargetVector::iterator it = targets.begin();
	TargetVector::iterator end = targets.end();
    
	for ( ; it != end; it++)
	{
		it->Discard();
	}
}

inline void Cloud::Discard( FeatureID id, double delta )
{
	TargetVector::iterator it = targets.begin();
	TargetVector::iterator end = targets.end();
    
	for ( ; it != end; it++)
	{
		it->Discard(id, delta);
	}
}


inline void Cloud::FindMinimumWeight( FeatureID id, double& min, int& dups)
{
	TargetVector::iterator it = targets.begin();
	TargetVector::iterator end = targets.end();
	for (it = targets.begin(); it != end; it++)
	{
		double wgt = it->WeightOfFeature(id);
		if ((min > wgt)) 
		{
			min = wgt;
			dups = 0;
		} else if (wgt == min)
		{
			// Two targets with the same weight
			// Don't discard yet
			dups++;
		} 
	}
}

inline void Cloud::CollectFeatures( set<FeatureID>& featureSet ) 
{
	TargetVector::iterator it = targets.begin();
	TargetVector::iterator end = targets.end();
	for ( ; it != end; it++)
	{
		it->CollectFeatures(featureSet);
	}
}

inline void Cloud::ResetCounters()
{
    TargetVector::iterator it = targets.begin();
    TargetVector::iterator end = targets.end();
    
    for ( ; it != end; it++)
    {
        it->ResetCounters();
    }
}

inline void Cloud::ShowStatistics( Counter total )
{
    TargetVector::iterator it = targets.begin();
    TargetVector::iterator end = targets.end();
    
    for ( ; it != end; it++)
    {
        it->ShowStatistics(total);
    }
}

inline void Cloud::Show()
{
    TargetVector::iterator it = targets.begin();
    TargetVector::iterator end = targets.end();
    
    for ( ; it != end; it++)
    {
        it->ShowFeatures();
    }
}

inline FeatureID Cloud::Id() const 
{
    return targetID;
}

#endif



